"""
Modified from the original implementation of CenterPoint in OpenPCDet.
Copyright [2020] [Shaoshuai.Shi] SPDX-License-Identifier: Apache-2.0(http://www.apache.org/licenses/LICENSE-2.0)
Modifications Copyright [2022] [LIVOX Perception Group]
"""

import copy
import torch
import numpy as np
import torch.nn as nn
from . import model_nms_utils

class SeparateHead(nn.Module):
    def __init__(self, input_channels, sep_head_dict, init_bias=-2.19, use_bias=False):
        super().__init__()
        self.sep_head_dict = sep_head_dict

        for cur_name in self.sep_head_dict:
            output_channels = self.sep_head_dict[cur_name]['out_channels']
            num_conv = self.sep_head_dict[cur_name]['num_conv']

            fc_list = []
            for k in range(num_conv - 1):
                fc_list.append(nn.Sequential(
                    nn.Conv2d(input_channels, input_channels, kernel_size=3, stride=1, padding=1, bias=use_bias),
                    nn.BatchNorm2d(input_channels),
                    nn.ReLU()
                ))
            fc_list.append(nn.Conv2d(input_channels, output_channels, kernel_size=3, stride=1, padding=1, bias=True))
            fc = nn.Sequential(*fc_list)
            if 'hm' in cur_name:
                fc[-1].bias.data.fill_(init_bias)
            else:
                for m in fc.modules():
                    if isinstance(m, nn.Conv2d):
                        nn.init.xavier_normal_(m.weight)
                        if hasattr(m, "bias") and m.bias is not None:
                            nn.init.constant_(m.bias, 0)

            self.__setattr__(cur_name, fc)

    def forward(self, x):
        ret_dict = {}
        for cur_name in self.sep_head_dict:
            ret_dict[cur_name] = self.__getattr__(cur_name)(x)

        return ret_dict


class CenterHead(nn.Module):
    def __init__(self, 
                 input_channels, 
                 num_class, 
                 class_names, 
                 point_cloud_range, 
                 voxel_size,
                 predict_boxes_when_training=True):
        super().__init__()
        self.num_class = num_class
        self.class_names = class_names
        self.point_cloud_range = point_cloud_range
        self.voxel_size = voxel_size

        # config
        self.class_name_branch_each_head = [class_names]

        self.separate_head_cfg = {
            'HEAD_ORDER': ['center', 'center_z', 'dim', 'rot'],
            'HEAD_DICT': {
                'center': {'out_channels':2, 'num_conv':1},
                'center_z': {'out_channels':1, 'num_conv':1},
                'dim': {'out_channels':3, 'num_conv':1},
                'rot': {'out_channels':2, 'num_conv':1}
            }
        }

        self.feature_map_stride = 1

        self.POST_PROCESSING = {
            'SCORE_THRESH': [0.2, 0.3, 0.3],
            'POST_CENTER_LIMIT_RANGE': self.point_cloud_range,
            'MAX_OBJ_PER_SAMPLE': 500,
            'NMS_CONFIG': {
                'NMS_TYPE': 'nms_gpu',
                'NMS_THRESH': 0.1,
                'NMS_PRE_MAXSIZE': 4096,
                'NMS_POST_MAXSIZE': 500
            }
        }

        self.class_names_each_head = []
        self.class_id_mapping_each_head = []
        for cur_class_names in self.class_name_branch_each_head:
            self.class_names_each_head.append([x for x in cur_class_names if x in class_names])
            cur_class_id_mapping = torch.from_numpy(np.array(
                [self.class_names.index(x) for x in cur_class_names if x in class_names]
            )).cuda()
            self.class_id_mapping_each_head.append(cur_class_id_mapping)
        
        total_classes = sum([len(x) for x in self.class_names_each_head])
        assert total_classes == len(self.class_names), f'class_names_each_head={self.class_names_each_head}'

        self.shared_conv = nn.Sequential(
            nn.Conv2d(
                input_channels, 64, 3, stride=1, padding=1,
                bias=True
            ),
            nn.BatchNorm2d(64),
            nn.ReLU(),
        )

        self.heads_list = nn.ModuleList()
        for idx, cur_class_names in enumerate(self.class_names_each_head):
            cur_head_dict = copy.deepcopy(self.separate_head_cfg['HEAD_DICT'])
            cur_head_dict['hm'] = dict(out_channels=len(cur_class_names), num_conv=1)
            self.heads_list.append(
                SeparateHead(
                    input_channels=64,
                    sep_head_dict=cur_head_dict,
                    init_bias=-2.19,
                    use_bias=True
                )
            )
        self.forward_ret_dict = {}

    @staticmethod
    def _gather_feat(feat, ind, mask=None):
        dim = feat.size(2)
        ind = ind.unsqueeze(2).expand(ind.size(0), ind.size(1), dim)
        feat = feat.gather(1, ind)
        if mask is not None:
            mask = mask.unsqueeze(2).expand_as(feat)
            feat = feat[mask]
            feat = feat.view(-1, dim)
        return feat

    @staticmethod
    def _transpose_and_gather_feat(feat, ind):
        feat = feat.permute(0, 2, 3, 1).contiguous()
        feat = feat.view(feat.size(0), -1, feat.size(3))
        feat = CenterHead._gather_feat(feat, ind)
        return feat

    @staticmethod
    def _topk(scores, K=40):
        batch, num_class, height, width = scores.size()

        topk_scores, topk_inds = torch.topk(scores.flatten(2, 3), K)

        topk_inds = topk_inds % (height * width)
        topk_ys = (topk_inds // width).float()
        topk_xs = (topk_inds % width).int().float()

        topk_score, topk_ind = torch.topk(topk_scores.view(batch, -1), K)
        topk_classes = (topk_ind // K).int()
        topk_inds = CenterHead._gather_feat(topk_inds.view(batch, -1, 1), topk_ind).view(batch, K)
        topk_ys = CenterHead._gather_feat(topk_ys.view(batch, -1, 1), topk_ind).view(batch, K)
        topk_xs = CenterHead._gather_feat(topk_xs.view(batch, -1, 1), topk_ind).view(batch, K)

        return topk_score, topk_inds, topk_classes, topk_ys, topk_xs

    @staticmethod
    def decode_bbox_from_heatmap(heatmap, rot_cos, rot_sin, center, center_z, dim,
                                 point_cloud_range=None, voxel_size=None, feature_map_stride=None, K=100,
                                 score_thresh=None, post_center_limit_range=None):
        batch_size, num_class, _, _ = heatmap.size()

        scores, inds, class_ids, ys, xs = CenterHead._topk(heatmap, K=K)
        center = CenterHead._transpose_and_gather_feat(center, inds).view(batch_size, K, 2)
        rot_sin = CenterHead._transpose_and_gather_feat(rot_sin, inds).view(batch_size, K, 1)
        rot_cos = CenterHead._transpose_and_gather_feat(rot_cos, inds).view(batch_size, K, 1)
        center_z = CenterHead._transpose_and_gather_feat(center_z, inds).view(batch_size, K, 1)
        dim = CenterHead._transpose_and_gather_feat(dim, inds).view(batch_size, K, 3)

        angle = torch.atan2(rot_sin, rot_cos)
        xs = xs.view(batch_size, K, 1) + center[:, :, 0:1]
        ys = ys.view(batch_size, K, 1) + center[:, :, 1:2]

        xs = xs * feature_map_stride * voxel_size[0] + point_cloud_range[0]
        ys = ys * feature_map_stride * voxel_size[1] + point_cloud_range[1]

        box_part_list = [xs, ys, center_z, dim, angle]

        final_box_preds = torch.cat((box_part_list), dim=-1)
        final_scores = scores.view(batch_size, K)
        final_class_ids = class_ids.view(batch_size, K)

        assert post_center_limit_range is not None
        mask = (final_box_preds[..., :3] >= post_center_limit_range[:3]).all(2)
        mask &= (final_box_preds[..., :3] <= post_center_limit_range[3:]).all(2)
        
        if score_thresh is not None:
            for i in range(len(score_thresh)):
                mask &= (~(final_class_ids == i)) | ((final_class_ids == i) & (final_scores > score_thresh[i])) 

        ret_pred_dicts = []
        for k in range(batch_size):
            cur_mask = mask[k]
            cur_boxes = final_box_preds[k, cur_mask]
            cur_scores = final_scores[k, cur_mask]
            cur_labels = final_class_ids[k, cur_mask]

            ret_pred_dicts.append({
                'pred_boxes': cur_boxes,
                'pred_scores': cur_scores,
                'pred_labels': cur_labels
            })
        return ret_pred_dicts

    def generate_predicted_boxes(self, batch_size, pred_dicts):
        post_process_cfg = self.POST_PROCESSING
        post_center_limit_range = torch.tensor(post_process_cfg['POST_CENTER_LIMIT_RANGE']).cuda().float()

        ret_dict = [{
            'pred_boxes': [],
            'pred_scores': [],
            'pred_labels': [],
        } for k in range(batch_size)]
        for idx, pred_dict in enumerate(pred_dicts):
            batch_hm = pred_dict['hm'].sigmoid()
            batch_center = pred_dict['center']
            batch_center_z = pred_dict['center_z']
            batch_dim = pred_dict['dim'].exp()
            batch_rot_cos = pred_dict['rot'][:, 0].unsqueeze(dim=1)
            batch_rot_sin = pred_dict['rot'][:, 1].unsqueeze(dim=1)
            batch_vel = pred_dict['vel'] if 'vel' in self.separate_head_cfg['HEAD_ORDER'] else None

            final_pred_dicts = CenterHead.decode_bbox_from_heatmap(
                heatmap=batch_hm, rot_cos=batch_rot_cos, rot_sin=batch_rot_sin,
                center=batch_center, center_z=batch_center_z, dim=batch_dim, 
                point_cloud_range=self.point_cloud_range, voxel_size=self.voxel_size,
                feature_map_stride=self.feature_map_stride,
                K=post_process_cfg['MAX_OBJ_PER_SAMPLE'],
                score_thresh=post_process_cfg['SCORE_THRESH'],
                post_center_limit_range=post_center_limit_range
            )
                
            for k, final_dict in enumerate(final_pred_dicts):
                final_dict['pred_labels'] = self.class_id_mapping_each_head[idx][final_dict['pred_labels'].long()]
                selected, selected_scores = model_nms_utils.class_agnostic_nms(
                    box_scores=final_dict['pred_scores'], box_preds=final_dict['pred_boxes'],
                    nms_config=post_process_cfg['NMS_CONFIG'],
                    score_thresh=None
                )

                final_dict['pred_boxes'] = final_dict['pred_boxes'][selected]
                final_dict['pred_scores'] = selected_scores
                final_dict['pred_labels'] = final_dict['pred_labels'][selected]

                ret_dict[k]['pred_boxes'].append(final_dict['pred_boxes'])
                ret_dict[k]['pred_scores'].append(final_dict['pred_scores'])
                ret_dict[k]['pred_labels'].append(final_dict['pred_labels'])

        for k in range(batch_size):
            ret_dict[k]['pred_boxes'] = torch.cat(ret_dict[k]['pred_boxes'], dim=0)
            ret_dict[k]['pred_scores'] = torch.cat(ret_dict[k]['pred_scores'], dim=0)
            ret_dict[k]['pred_labels'] = torch.cat(ret_dict[k]['pred_labels'], dim=0) + 1

        return ret_dict

    def forward(self, data_dict):
        spatial_features_2d = data_dict['spatial_features_2d']
        x = self.shared_conv(spatial_features_2d)

        pred_dicts = []
        for head in self.heads_list:
            pred_dicts.append(head(x))
        
        self.forward_ret_dict['pred_dicts'] = pred_dicts

        pred_dicts = self.generate_predicted_boxes(
            data_dict['batch_size'], pred_dicts
        )

        data_dict['final_box_dicts'] = pred_dicts
        return data_dict



import numpy as np

class PointCloud:
    """
    General PointCloud class.
    Holds point cloud data as an (N, 4) numpy array: [x, y, z, intensity]
    """
    def __init__(self, points: np.ndarray):
        assert isinstance(points, np.ndarray), "points must be a numpy array"
        assert points.shape[1] == 4, "PointCloud expects shape (N, 4)"
        self.points = points.astype(np.float32)


import copy
import torch
import torch.nn as nn

class BoolMap(nn.Module):
    '''
             ^ X,W
             |
             |
     Y,H     |
    <--------o
    '''
    def __init__(self, point_cloud_range,voxel_size=[0.2,0.2,0.2], **kwargs):
        super().__init__()
        # VOXEL SIZE
        self.DX, self.DY, self.DZ = \
            voxel_size[0], voxel_size[1],voxel_size[2]
        
        # ROI in meter
        self.m_x_min = point_cloud_range[0]
        self.m_x_max = point_cloud_range[3]

        self.m_y_min = point_cloud_range[1]
        self.m_y_max = point_cloud_range[4]

        self.m_z_min = point_cloud_range[2]
        self.m_z_max = point_cloud_range[5]

        # SIZE of BEV map
        self.BEV_W = round((point_cloud_range[3]-point_cloud_range[0])/self.DX)
        self.BEV_H = round((point_cloud_range[4]-point_cloud_range[1])/self.DY)
        self.BEV_C = round((point_cloud_range[5]-point_cloud_range[2])/self.DZ)

        self.num_bev_features = self.BEV_C
        

    def forward(self, batch_dict):
        """
        Args:
            data_dict:
                points: (num_points, 1+4)

        Returns:
            batch_dict:
                batch_dict['spatial_features'] = spatial_features (B,C,H,W)

        """
        pc_lidar = batch_dict['points'].clone() 

        bev_img = torch.cuda.BoolTensor(batch_dict['batch_size'],self.BEV_C,self.BEV_H,self.BEV_W).fill_(0)
        pc_lidar[:,1]=((pc_lidar[:,1]-self.m_x_min)/self.DX)
        pc_lidar[:,2]=((pc_lidar[:,2]-self.m_y_min)/self.DY)
        pc_lidar[:,3]=((pc_lidar[:,3]-self.m_z_min)/self.DZ)
        pc_lidar = pc_lidar.trunc().long()
        bev_img[pc_lidar[:,0], pc_lidar[:,3], pc_lidar[:,2], pc_lidar[:,1]] = 1
        bev_img = bev_img.float() 
        batch_dict['spatial_features'] = bev_img

        return batch_dict



import torch
import numpy as np

# Convert outputs to CPU
def move_to_cpu(item):
    if isinstance(item, torch.Tensor):
        return item.cpu().numpy()
    elif isinstance(item, dict):
        return {k: move_to_cpu(v) for k, v in item.items()}
    elif isinstance(item, list):
        return [move_to_cpu(i) for i in item]
    else:
        return item
    

def load_data_to_gpu(batch_dict:dict):
    for key, val in batch_dict.items():
        if isinstance(val, np.ndarray):
            batch_dict[key] = torch.from_numpy(val).float().cuda()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F

from .boolmap import BoolMap
from .resfpn import ResBEVBackboneConcat
from .centerhead import CenterHead

class LD_base(nn.Module):
    def __init__(self):
        super(LD_base, self).__init__()
        self.voxel_size = [0.2, 0.2, 0.2]
        self.point_cloud_range = [0, -44.8, -2, 224, 44.8, 4] 
        self.point_to_bev = BoolMap(self.point_cloud_range, voxel_size=self.voxel_size)
        self.backbone = ResBEVBackboneConcat(30)
        self.head = CenterHead(input_channels=128, 
                               num_class=3, 
                               class_names=['Vehicle', 'Pedestrian', 'Cyclist'], 
                               point_cloud_range=self.point_cloud_range, 
                               voxel_size=self.voxel_size)
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
    
    def forward(self, batch_dict):
        batch_dict = self.point_to_bev(batch_dict)
        batch_dict = self.backbone(batch_dict)
        batch_dict = self.head(batch_dict)
        return batch_dict['final_box_dicts']




import copy
import numpy as np
import torch
from collections import defaultdict
from ..types.PointCloud import PointCloud
from ..utils import move_to_cpu, load_data_to_gpu

def mask_points_out_of_range(pc, pc_range):
    pc_range = np.array(pc_range)
    pc_range[3:6] -= 0.01  # Avoid boundary issues
    mask_x = (pc[:, 0] > pc_range[0]) & (pc[:, 0] < pc_range[3])
    mask_y = (pc[:, 1] > pc_range[1]) & (pc[:, 1] < pc_range[4])
    mask_z = (pc[:, 2] > pc_range[2]) & (pc[:, 2] < pc_range[5])
    return pc[mask_x & mask_y & mask_z]

def collate_batch(batch_list):
    data_dict = defaultdict(list)
    for sample in batch_list:
        for key, val in sample.items():
            data_dict[key].append(val)
    ret = {}
    for key, val in data_dict.items():
        if key == 'points':
            coors = []
            for i, coor in enumerate(val):
                coor_pad = np.pad(coor, ((0, 0), (1, 0)), mode='constant', constant_values=i)
                coors.append(coor_pad)
            ret[key] = np.concatenate(coors, axis=0)
    ret['batch_size'] = len(batch_list)
    return ret

def infer_pointcloud_detection(model: torch.nn.Module, pointcloud: PointCloud) -> dict:
    """
    Perform inference on general point cloud using the provided model.

    Args:
        model (torch.nn.Module): The 3D object detection model (e.g., LD_base).
        pointcloud (PointCloud): Input point cloud with shape (N, 4) [x, y, z, intensity]

    Returns:
        dict: The prediction dictionary output by the model (moved to CPU).
    """
    offset_angle = 0.0  # degrees
    offset_ground = 1.8  # meters
    point_cloud_range = [0, -44.8, -2, 224, 44.8, 4]

    points_list = copy.deepcopy(pointcloud.points)

    # Preprocess
    points_list[:, 2] += points_list[:, 0] * np.tan(offset_angle / 180. * np.pi) + offset_ground
    points_list = mask_points_out_of_range(points_list, point_cloud_range)

    input_dict = {
        'points': points_list
    }

    data_batch = collate_batch([input_dict])
    load_data_to_gpu(data_batch)

    model.eval()
    with torch.no_grad():
        starter, ender = torch.cuda.Event(enable_timing=True), torch.cuda.Event(enable_timing=True)
        torch.cuda.synchronize()
        starter.record()
        pred_dicts = model.forward(data_batch)
        ender.record()
        torch.cuda.synchronize()
        elapsed_time_ms = starter.elapsed_time(ender)
        print(f"Inference time: {elapsed_time_ms:.2f} ms")


    pred_dicts = move_to_cpu(pred_dicts)

    return pred_dicts


"""
Copyright [2020] [Shaoshuai.Shi] SPDX-License-Identifier: Apache-2.0(http://www.apache.org/licenses/LICENSE-2.0)
"""
import torch
from ...ops.iou3d_nms import iou3d_nms_utils

def class_agnostic_nms(box_scores, box_preds, nms_config, score_thresh=None):
    src_box_scores = box_scores
    if score_thresh is not None:
        scores_mask = (box_scores >= score_thresh)
        box_scores = box_scores[scores_mask]
        box_preds = box_preds[scores_mask]

    selected = []
    if box_scores.shape[0] > 0:
        box_scores_nms, indices = torch.topk(box_scores, k=min(nms_config['NMS_PRE_MAXSIZE'], box_scores.shape[0]))
        boxes_for_nms = box_preds[indices]
        keep_idx, selected_scores = getattr(iou3d_nms_utils, nms_config['NMS_TYPE'])(
                boxes_for_nms[:, 0:7], box_scores_nms, nms_config['NMS_THRESH'], **nms_config
        )
        selected = indices[keep_idx[:nms_config['NMS_POST_MAXSIZE']]]

    if score_thresh is not None:
        original_idxs = scores_mask.nonzero().view(-1)
        selected = original_idxs[selected]
    return selected, src_box_scores[selected]


def multi_classes_nms(cls_scores, box_preds, nms_config, score_thresh=None):
    """
    Args:
        cls_scores: (N, num_class)
        box_preds: (N, 7 + C)
        nms_config:
        score_thresh:

    Returns:

    """
    pred_scores, pred_labels, pred_boxes = [], [], []
    for k in range(cls_scores.shape[1]):
        if score_thresh is not None:
            scores_mask = (cls_scores[:, k] >= score_thresh)
            box_scores = cls_scores[scores_mask, k]
            cur_box_preds = box_preds[scores_mask]
        else:
            box_scores = cls_scores[:, k]
            cur_box_preds = box_preds

        selected = []
        if box_scores.shape[0] > 0:
            box_scores_nms, indices = torch.topk(box_scores, k=min(nms_config['NMS_PRE_MAXSIZE'], box_scores.shape[0]))
            boxes_for_nms = cur_box_preds[indices]
            keep_idx, selected_scores = getattr(iou3d_nms_utils, nms_config['NMS_TYPE'])(
                    boxes_for_nms[:, 0:7], box_scores_nms, nms_config['NMS_THRESH'], **nms_config
            )
            selected = indices[keep_idx[:nms_config['NMS_POST_MAXSIZE']]]

        pred_scores.append(box_scores[selected])
        pred_labels.append(box_scores.new_ones(len(selected)).long() * k)
        pred_boxes.append(cur_box_preds[selected])

    pred_scores = torch.cat(pred_scores, dim=0)
    pred_labels = torch.cat(pred_labels, dim=0)
    pred_boxes = torch.cat(pred_boxes, dim=0)

    return pred_scores, pred_labels, pred_boxes



from setuptools import setup, find_packages
from torch.utils.cpp_extension import BuildExtension, CUDAExtension
import os

def make_cuda_ext(name, module, sources):
    cuda_ext = CUDAExtension(
        name='%s.%s' % (module, name),
        sources=[os.path.join(*module.split('.'), src) for src in sources]
    )
    return cuda_ext

with open("readme.md", "r") as readme_file:
    readme = readme_file.read()

requirements = [
    "torch==1.8.0",
    'numpy==1.19.2',
] 

setup(
    name="det3d",
    version="0.0.5",
    author="Tao Xiang",
    author_email="xiang.tao@outlook.de",
    description="A package of 3D object detection models",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/leoxiang66/3D-Detection",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "det3d.ops.iou3d_nms": ["src/*.cpp", "src/*.cu", "src/*.h"],
    },
    install_requires=requirements,
    cmdclass={
        'build_ext': BuildExtension,  # 用于编译 CUDA 扩展
    },
    ext_modules=[
        make_cuda_ext(
            name='iou3d_nms_cuda',
            module='det3d.ops.iou3d_nms',
            sources=[
                'src/iou3d_cpu.cpp',
                'src/iou3d_nms_api.cpp',
                'src/iou3d_nms.cpp',
                'src/iou3d_nms_kernel.cu',
            ],
        ),
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F

class BasicBlock(nn.Module):
    """Basic Block for resnet 18 and resnet 34
    """
    #BasicBlock and BottleNeck block
    #have different output size
    #we use class attribute expansion
    #to distinct
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()

        #residual function
        self.residual_function = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels * BasicBlock.expansion, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels * BasicBlock.expansion),
            nn.ReLU(inplace=True)
        )

        #shortcut
        self.shortcut = nn.Sequential()

        #the shortcut output dimension is not the same with residual function
        #use 1*1 convolution to match the dimension
        if stride != 1 or in_channels != BasicBlock.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels * BasicBlock.expansion, kernel_size=1, stride=stride),
                nn.BatchNorm2d(out_channels * BasicBlock.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))

class BottleNeck(nn.Module):
    """Residual block for resnet over 50 layers
    """
    expansion = 1
    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.residual_function = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, stride=stride, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels * BottleNeck.expansion, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels * BottleNeck.expansion),
        )

        self.shortcut = nn.Sequential()

        if stride != 1 or in_channels != out_channels * BottleNeck.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels * BottleNeck.expansion, stride=stride, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels * BottleNeck.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))


class ResBEVBackboneConcat(nn.Module):
    """
    Modified from the original implementation of BEV backbone in OpenPCDet.
    """  
    def __init__(self, 
                 input_channels,
                 layer_nums=[2, 2, 3, 3, 2], 
                 layer_strides=[2, 2, 2, 2, 2], 
                 num_filters=[32, 48, 64, 96, 128], 
                 upsample_strides=[2, 4, 8, 16, 32]):
        super().__init__()
        num_levels = len(layer_nums)
        c_in_list = [input_channels, *num_filters[:-1]]
        self.blocks = nn.ModuleList()
        self.deblocks = nn.ModuleList()
        for idx in range(num_levels):
            cur_layers = [
                nn.ZeroPad2d(1),
                nn.Conv2d(
                    c_in_list[idx], num_filters[idx], kernel_size=3,
                    stride=layer_strides[idx], padding=0, bias=False
                ),
                nn.BatchNorm2d(num_filters[idx], eps=1e-3, momentum=0.01),
                nn.ReLU()
            ]
            for k in range(layer_nums[idx]):
                cur_layers.extend([BottleNeck(num_filters[idx], num_filters[idx])])
            self.blocks.append(nn.Sequential(*cur_layers))

            if len(upsample_strides) > 0:
                self.deblocks.append(nn.Sequential(
                    nn.UpsamplingBilinear2d(scale_factor=upsample_strides[idx]),
                ))

        self.fushion = nn.Sequential(
            nn.Conv2d(sum(num_filters), 128, kernel_size=1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True)
        )

        self.attention_w = nn.Sequential(
            nn.Conv2d(128, 128, kernel_size=1, bias=False),
            nn.BatchNorm2d(128),
        )

        self.num_bev_features = 128

    def forward(self, batch_dict):
        spatial_features = batch_dict['spatial_features']
        ups = []
        ret_dict = {}
        x = spatial_features
        for i in range(len(self.blocks)):
            x = self.blocks[i](x)

            stride = int(spatial_features.shape[2] / x.shape[2])
            ret_dict['spatial_features_%dx' % stride] = x
            if len(self.deblocks) > 0:
                ups.append(self.deblocks[i](x))
            else:
                ups.append(x)

        x = torch.cat(ups, dim=1)
        x = self.fushion(x)
        w_x = torch.softmax(self.attention_w(x), dim=1)
        x = w_x * x

        batch_dict['spatial_features_2d'] = x
        return batch_dict





"""
3D IoU Calculation and Rotated NMS
Written by Shaoshuai Shi
All Rights Reserved 2019-2020.
"""
import torch
import numpy as np
from . import iou3d_nms_cuda

def check_numpy_to_torch(x):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).float(), True
    return x, False

def boxes_bev_iou_cpu(boxes_a, boxes_b):
    """
    Args:
        boxes_a: (N, 7) [x, y, z, dx, dy, dz, heading]
        boxes_b: (M, 7) [x, y, z, dx, dy, dz, heading]

    Returns:
        ans_iou: (N, M)
    """
    boxes_a, is_numpy = check_numpy_to_torch(boxes_a)
    boxes_b, is_numpy = check_numpy_to_torch(boxes_b)
    assert not (boxes_a.is_cuda or boxes_b.is_cuda), 'Only support CPU tensors'
    assert boxes_a.shape[1] == 7 and boxes_b.shape[1] == 7
    ans_iou = boxes_a.new_zeros(torch.Size((boxes_a.shape[0], boxes_b.shape[0])))
    iou3d_nms_cuda.boxes_iou_bev_cpu(boxes_a.contiguous(), boxes_b.contiguous(), ans_iou)

    return ans_iou.numpy() if is_numpy else ans_iou


def boxes_iou_bev(boxes_a, boxes_b):
    """
    Args:
        boxes_a: (N, 7) [x, y, z, dx, dy, dz, heading]
        boxes_b: (M, 7) [x, y, z, dx, dy, dz, heading]

    Returns:
        ans_iou: (N, M)
    """
    assert boxes_a.shape[1] == boxes_b.shape[1] == 7
    ans_iou = torch.cuda.FloatTensor(torch.Size((boxes_a.shape[0], boxes_b.shape[0]))).zero_()

    iou3d_nms_cuda.boxes_iou_bev_gpu(boxes_a.contiguous(), boxes_b.contiguous(), ans_iou)

    return ans_iou


def boxes_iou3d_gpu(boxes_a, boxes_b):
    """
    Args:
        boxes_a: (N, 7) [x, y, z, dx, dy, dz, heading]
        boxes_b: (M, 7) [x, y, z, dx, dy, dz, heading]

    Returns:
        ans_iou: (N, M)
    """
    assert boxes_a.shape[1] == boxes_b.shape[1] == 7

    # height overlap
    boxes_a_height_max = (boxes_a[:, 2] + boxes_a[:, 5] / 2).view(-1, 1)
    boxes_a_height_min = (boxes_a[:, 2] - boxes_a[:, 5] / 2).view(-1, 1)
    boxes_b_height_max = (boxes_b[:, 2] + boxes_b[:, 5] / 2).view(1, -1)
    boxes_b_height_min = (boxes_b[:, 2] - boxes_b[:, 5] / 2).view(1, -1)

    # bev overlap
    overlaps_bev = torch.cuda.FloatTensor(torch.Size((boxes_a.shape[0], boxes_b.shape[0]))).zero_()  # (N, M)
    iou3d_nms_cuda.boxes_overlap_bev_gpu(boxes_a.contiguous(), boxes_b.contiguous(), overlaps_bev)

    max_of_min = torch.max(boxes_a_height_min, boxes_b_height_min)
    min_of_max = torch.min(boxes_a_height_max, boxes_b_height_max)
    overlaps_h = torch.clamp(min_of_max - max_of_min, min=0)

    # 3d iou
    overlaps_3d = overlaps_bev * overlaps_h

    vol_a = (boxes_a[:, 3] * boxes_a[:, 4] * boxes_a[:, 5]).view(-1, 1)
    vol_b = (boxes_b[:, 3] * boxes_b[:, 4] * boxes_b[:, 5]).view(1, -1)

    iou3d = overlaps_3d / torch.clamp(vol_a + vol_b - overlaps_3d, min=1e-6)

    return iou3d


def nms_gpu(boxes, scores, thresh, pre_maxsize=None, **kwargs):
    """
    :param boxes: (N, 7) [x, y, z, dx, dy, dz, heading]
    :param scores: (N)
    :param thresh:
    :return:
    """
    assert boxes.shape[1] == 7
    order = scores.sort(0, descending=True)[1]
    if pre_maxsize is not None:
        order = order[:pre_maxsize]

    boxes = boxes[order].contiguous()
    keep = torch.LongTensor(boxes.size(0))
    num_out = iou3d_nms_cuda.nms_gpu(boxes, keep, thresh)
    return order[keep[:num_out].cuda()].contiguous(), None


def nms_normal_gpu(boxes, scores, thresh, **kwargs):
    """
    :param boxes: (N, 7) [x, y, z, dx, dy, dz, heading]
    :param scores: (N)
    :param thresh:
    :return:
    """
    assert boxes.shape[1] == 7
    order = scores.sort(0, descending=True)[1]

    boxes = boxes[order].contiguous()

    keep = torch.LongTensor(boxes.size(0))
    num_out = iou3d_nms_cuda.nms_normal_gpu(boxes, keep, thresh)
    return order[keep[:num_out].cuda()].contiguous(), None
